#include <avr/io.h>
#include <stdio.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define F_CPU 8000000  //ta sama definicje nalezy zrobic w opcjach tolchain
#define timerinit 230
#define timeinterval  10 //funkcja wyznaczajaca minimalna wartosc czasu pomiedzy zmianami
#define usebuttons 1     //nalezy wpisac 1 w przypadku zalaczenia uzywania przyciskow zamiast bluetoth
#define timerupdwn    150;
#define timetoeeprom  15000;

#define BAUD 9600      //predkosc 9600
#define BAUD_PRESCALE   (((F_CPU / (BAUD * 16UL))) - 1)

//--------------------------------------------------------------------------------
//Deklaracje dla sterowania portami
#define onpowercloset  PORTB|=(1<<PORTB0)
#define onpowerfoot1   PORTB|=(1<<PORTB1)
#define onpowerfoot2   PORTB|=(1<<PORTB1)
#define onpowerstrip   PORTB|=(1<<PORTB2)

#define offpowercloset  PORTB&=~(1<<PORTB0)
#define offpowerfoot1   PORTB&=~(1<<PORTB1)
#define offpowerfoot2   PORTB&=~(1<<PORTB1)
#define offpowerstrip   PORTB&=~(1<<PORTB2)

#define onsr1           PORTC|=(1<<PORTC0)
#define onsg1           PORTC|=(1<<PORTC1)
#define onsb1           PORTC|=(1<<PORTC2)

#define onfootr1        PORTC|=(1<<PORTC3)
#define onfootg1        PORTC|=(1<<PORTC4)
#define onfootb1        PORTC|=(1<<PORTC5)

#define onclosetr1      PORTD|=(1<<PORTD7)
#define onclosetg1      PORTD|=(1<<PORTD6)
#define onclosetb1      PORTD|=(1<<PORTD5)

#define offsr1           PORTC&=~(1<<PORTC0)
#define offsg1           PORTC&=~(1<<PORTC1)
#define offsb1           PORTC&=~(1<<PORTC2)

#define offfootr1        PORTC&=~(1<<PORTC3)
#define offfootg1        PORTC&=~(1<<PORTC4)
#define offfootb1        PORTC&=~(1<<PORTC5)

#define offclosetr1      PORTD&=~(1<<PORTD7)
#define offclosetg1      PORTD&=~(1<<PORTD6)
#define offclosetb1      PORTD&=~(1<<PORTD5)

#if usebuttons==1
#define	wup     !(PIND & (1<<PORTD2))
#define	wselect !(PIND & (1<<PORTD3))
#define	wdown   !(PIND & (1<<PORTD4))
#endif

//********************************************************************************
//Deklaracje funkcji
void UART_init(void);
void Timer_init(void);
void PORT_init(void);
void Buffor_Searching(void);
void Change_value(void);
//Obsluga pamieci EEPROM
unsigned char EEPROM_read(unsigned int uiAddress);
void EEPROM_write(unsigned int uiAddress, unsigned char ucData);
void Write_EEPROM(void);
void Read_EEPROM(void);
#if usebuttons==1
void Init_Buttons(void);
void Control_Buttons(void);
#endif

//*********************************************************************************
//Zmienne funkcji
volatile uint16_t timerselect, timerup, timerdwn, timerforeep;
volatile uint8_t counter, wskbuff, rxflag, timerus, flagcount, counter2;
volatile uint8_t functions[3][4];
unsigned char buffor[100];
uint8_t tempbuffor[3][4]={{0, 0, 0, 0},{0xff, 0xff, 0, 0}, {0, 0, 0, 0}};  //wartosci poczatkowe dla wszystkich PWM
uint8_t firstrun=1;  //zabezpieczenie przed nadmierny zapisem do eeprom podczas uruchamiania procesora


int main(void)
{
    
	 
	 Timer_init(); //inicjalizacja timera
	 PORT_init();  //inicjalizacja portow
	 
	 UART_init(); //inicjalizacja ukladu UART
	 
	 #if usebuttons==1
	 Init_Buttons();  //inicjalizacja przyciskow
	 #endif
	 
	 sei();  //wlaczenie przerwan globalnych
	 
	//Pierwsze wejscie do zapisu pamieci eeprom
	if(0x2a!=EEPROM_read(0))
	{
		EEPROM_write(0, 0x2a); //wpisanie wartosci blokujacej nadpisywanie do pamieci EEPROM
		Write_EEPROM(); //wpisanie aktualnej wartosci
	}
	Read_EEPROM();

	
    while (1) 
    {
		
			if(rxflag) //zezwolenie na przeszkukanie bufora odbiorczego w poszukiwaniu danych
			{
				Buffor_Searching(); //przeszukanie buffora odbiorczego
				rxflag=0;
			}
			
			Change_value(); //procedura zmieniajaca ustawione wartosci w czasie
			
			#if usebuttons==1
			 Control_Buttons(); //funkcja obslugujaca przyciski
			#endif
		
    }
}

void Timer_init(void)  //inicjalizacja timera
{
	TIMSK0 |= (1<<TOIE0);            //Przerwanie od przepelnienia - overflow
	TCCR0B |= (1<<CS01);            // rdem CLK, preskaler 8
	TCNT0 = timerinit;              //Pocztkowa warto licznika
}

void UART_init(void)  //inicjalizacja UART
{
	
	UBRR0H = (BAUD_PRESCALE >> 8);        //wpisanie starszego bajtu
	UBRR0L = BAUD_PRESCALE;             //wpisanie mlodszego bajtu

	
	UCSR0C = (0<<UMSEL00) | (0<<UMSEL01) | (1<<UCSZ00) | (1<<UCSZ01) | (0<<UPM00) |  (0<<UPM01) | (0<<USBS0);  //8 bitw, 1 bit stopu, parzystosc brak,
	UCSR0B = (1<<TXEN0) | (1<<RXEN0) | (1<<RXCIE0); //wcz  odbiornik oraz ich przerwania odbiornik; //wcz  odbiornik oraz ich przerwania odbiornik	
}

void PORT_init(void)  //inicjalizacja PORTow
{
	//Ustawienie kierunkow dla obslugiwanych portow, wejscie, wyjscie
	DDRB=0x07; 
	DDRD=0xE3;
	DDRC=0x3f;
	
	PORTB=0x00;
	PORTD=0x00;
	PORTC=0x00;
}

void Buffor_Searching(void)  //przeszukanie buffora odbiorczego
{
	uint8_t i, j, k, l, m, n, o, p;
	uint8_t nloop;
	uint8_t temptemp;
	
	
	for(nloop=0; nloop<100; nloop++)
	{
		//inkrementacja wskaznikow na zmienna znakowa
		i=nloop;
		j=i+1;
		k=i+2;
		l=i+3;
		m=i+4;
		n=i+5;
		o=i+6;
		p=i+7;
		
		//jezeli przekraca 99 to ustawienie wskaznikow na poczatek bufora
		if(j>=100) j=j-100;
		if(k>=100) k=k-100;
		if(l>=100) l=l-100;
		if(m>=100) m=m-100;
		if(n>=100) n=n-100;
		if(o>=100) o=o-100;
		if(p>=100) p=p-100;
		
		
		
		if(buffor[i]=='s' && buffor[j]==':' && buffor[p]=='k')
		{
			temptemp=buffor[k]-0x30;
			if(temptemp<3)
			{
				tempbuffor[temptemp][0]=(uint8_t)buffor[l];
				tempbuffor[temptemp][1]=(uint8_t)buffor[m];
				tempbuffor[temptemp][2]=(uint8_t)buffor[n];
				tempbuffor[temptemp][3]=(uint8_t)buffor[o];
				
				//wyczyszczenie miejsca buffora jesli dane zostaly odebrane
				buffor[k]=0x00;
				buffor[l]=0x00;
				buffor[m]=0x00;
				buffor[n]=0x00;
				buffor[o]=0x00;
				buffor[p]=0x00;
				buffor[i]=0x00;
				buffor[j]=0x00;
			}			
		}
			
		
	}
	

	
}

void Change_value(void) //funkcja zmieniajaca stopniowo wartosci
{
	uint8_t k, j;
	static uint8_t flagazapisu=0;
	uint8_t differ;
		
		if(!timerus)
		{
			for(k=0; k<3; k++)
			{
				for(j=0; j<4; j++)
				{
					if(functions[k][j]!=tempbuffor[k][j])
					{
						if(functions[k][j]<tempbuffor[k][j])
						{
							differ=tempbuffor[k][j]-functions[k][j];
							if(differ>=4) functions[k][j]+=4; else
							functions[k][j]+=1;
						}		
						else
						{
							differ=functions[k][j]-tempbuffor[k][j];
							if(differ>=4) functions[k][j]-=4; else
							 functions[k][j]-=1; 
						}
						flagazapisu=1;  //zezwolenie na zapis do pamieci eeprom
					}
					else if(flagazapisu)
					{
						if(!firstrun)
						{
							if(!timerforeep) Write_EEPROM();   //jezeli od nacisniecia przycisku nastapila chwila
					    } 
						else  firstrun=0;  //wylaczenie przed nadmiernym zapisem do pamieci eeprom podczas uruchamiania procesora	   
						flagazapisu=0; //wykasowanie flagi zezwalajacej na zapis do pamieci eeprom
						
					}
				}		
			}
			
			timerus=timeinterval;
		}
		
		
}

//--------------------------------------------------------------------------------------------
ISR(USART_RX_vect )//przerwania od odbioru UART
{
	buffor[wskbuff]=UDR0;
	
	wskbuff++;
	if(wskbuff>99) wskbuff=0;
	rxflag=1;  //zezwolenie na przeszukanie bufora z danymi
	
}


ISR(TIMER0_OVF_vect) // przerwanie od przepelnienia TIMER0
{
	
/*------------------------------------------------------------------------------
//Uwaga w przypadku widocznych zjawisk optycznych nalezy zwiekszyc inkrementacje
//inktementacje lincznikow counter+=2 counter2+=4
------------------------------------------------------------------------------*/
	counter++; //przeladowanie i zwiekszanie dla PWM
	counter2+=3;
	
	
	if(counter2>functions[0][0]) offpowercloset; else onpowercloset;
	if(counter2>functions[1][0]) offpowerfoot1; else onpowerfoot1;
    if(counter2>functions[1][0]) offpowerfoot2; else onpowerfoot2;
    if(counter2>functions[2][0]) offpowerstrip; else onpowerstrip;
	
	if(counter>functions[0][1]) offclosetr1; else onclosetr1;
	if(counter>functions[0][2]) offclosetg1; else onclosetg1;
	if(counter>functions[0][3]) offclosetb1; else onclosetb1;
	
	if(counter>functions[1][1]) offfootr1; else onfootr1;
	if(counter>functions[1][2]) offfootg1; else onfootg1;
	if(counter>functions[1][3]) offfootb1; else onfootb1;
	
	if(counter>functions[2][1]) offsr1; else onsr1;
	if(counter>functions[2][2]) offsg1; else onsg1;
	if(counter>functions[2][3]) offsb1; else onsb1;
	
	if(timerus) timerus--; //zmienna do odliczania czasu
	if(timerselect) timerselect--;
	if(timerup) timerup--;   //timery dla inkrementacji lub dekrementacji wartosci wybranej
	if(timerdwn) timerdwn--;
	if(timerforeep) timerforeep--;
	
	TCNT0 = timerinit;  //ponowne wpisanie wartosci poczatkowej
}


//-----------------------------------------------------------------
//kontrola obslugi za pomoca przyciskow
#if usebuttons==1
void Init_Buttons(void)
{
	DDRD&=~((1<<PORTD2) | (1<<PORTD3) | (1<<PORTD4)); //ustawienie portow wejsciowych
	PORTD|=((1<<PORTD2) | (1<<PORTD3) | (1<<PORTD4)); //podniesienie w stan wysoki wszystkich portow
}
void Control_Buttons(void)
{
	static uint8_t drgsel=0;
	static uint8_t selsel=0;
	static uint8_t xsel=0, ysel=0;
	static uint8_t flaga2=0;
	static uint8_t flag=0;
	
	
	if(wselect && !drgsel)
	{
		drgsel=200; //elyminacja drgan
		selsel++;
		if(selsel>11) selsel=0;
		xsel=selsel/4;
		ysel=selsel%4;
		
		timerselect=15000; //zezwolenie na pokazanie ktora mamy wybrana opcje
		timerforeep=timetoeeprom;
	}
	else if(!wselect && drgsel) drgsel--;  //dekrementacja zmienne elm. drgania
	
	//Inkrementacja lub dekrementacja wybranej wartosci
		if(wup && !timerup)
		{
			timerup=timerupdwn;
			if(tempbuffor[xsel][ysel]<255) tempbuffor[xsel][ysel]+=1;
			timerforeep=timetoeeprom;  //czas musi byc wiekszy niz czas timerupdown
			flag=1;
		}
		
		if(wdown && !timerdwn)
		{
			timerdwn=timerupdwn;
			if(tempbuffor[xsel][ysel]>0) tempbuffor[xsel][ysel]-=1;
			timerforeep=timetoeeprom;  //czas musi byc wiekszy niz czas timerupdown
			flag=1;
		}
		
		if(timerselect) //jezeli mamy zezwolnie na pokazanie ktora opcje mamy wybrana
		{
			flaga2=1; //zezwolenie na pozniejszy powrot do ustawionych wartosci bez lagodnego przejscia
			
			for(uint8_t i=0; i<4; i++)
			{
				if(ysel==i)
				{
					if(ysel>0)
					{
					 functions[xsel][ysel]=253;
					 functions[xsel][0]=253;
					}
					else
					{
						functions[xsel][0]=253;
						functions[xsel][1]=253;
						functions[xsel][2]=253;
						functions[xsel][3]=253;
						break; //wyjscie z petli
					}
				}
				else if(i>0)
				{
					functions[xsel][i]=0;
				}
				
			}		
		}
		
		if(!timerselect && flaga2)
		{
			flaga2=0; //wykasowanie zezwolenia na wejscie bez animacji do ustawionych wartosci po mignieciu oznaczajacym wybor ustawienia
			
			for(uint8_t s=0; s<3; s++)
			{
				for(uint8_t p=0; p<4; p++)
				{
					functions[s][p]=tempbuffor[s][p];
				}
			}
			
		}
		
		if(!timerforeep && flag)
		{
			flag=0;
			 Write_EEPROM();   //jezeli od nacisniecia przycisku nastapila chwila i zmieniono wartosc to zapis do eeprom
		}
	
}
#endif

//---------------------------------------------------------------
//Obsuga pamici EEPROM
void EEPROM_write(unsigned int uiAddress, unsigned char ucData)
{

	while(EECR & (1<<EEPE));
	EEARH = (uint8_t)(uiAddress>>8);
	EEARL = (uint8_t)uiAddress;
	EEDR = ucData;
	EECR |= (1<<EEMPE);
	EECR |= (1<<EEPE);
}

unsigned char EEPROM_read(unsigned int uiAddress)
{
	while(EECR & (1<<EEPE));
	EEARH = (uint8_t)(uiAddress>>8);
	EEARL = (uint8_t)uiAddress;
	EECR |= (1<<EERE);
	return EEDR;
}

void Write_EEPROM(void)
{
uint8_t i, k, eep;

	eep=1;
	for(i=0; i<3; i++)
	{
		for(k=0; k<4; k++)
		{
		EEPROM_write(eep, tempbuffor[i][k]);
		eep++;
		}
	}	
	
}

void Read_EEPROM(void)
{
uint8_t i, k, eep;

	eep=1;
	for(i=0; i<3; i++)
	{
		for(k=0; k<4; k++)
		{
			tempbuffor[i][k]=EEPROM_read(eep);
			eep++;
		}
	}	
}